@@ -1,11 +1,11 @@
 # noinspection PyPep8
 
 import shutil
-import sys
-
 from .fmi1 import *
 from .fmi1 import _FMU1
 from .fmi2 import *
+from .fmi2 import _FMU2
+from . import fmi3
 from . import extract
 from .util import auto_interval
 import numpy as np
@@ -30,17 +30,10 @@ class Recorder(object):
         self.fmu = fmu
         self.interval = interval
 
-        self.cols = [('time', np.float64)]
+        self.cols = [('time', np.float64, None)]  # name, dtype, shape
         self.rows = []
-
-        real_names = []
-        self.real_vrs = []
-
-        integer_names = []
-        self.integer_vrs = []
-
-        boolean_names = []
-        self.boolean_vrs = []
+        self.info = {}  # type -> (names, vrs, shapes, n_values, getter)
+        self.types = []
 
         self.constants = {}
         self.modelDescription = modelDescription
@@ -50,22 +43,44 @@ class Recorder(object):
 
             # collect the variables to record
             if (variableNames is not None and sv.name in variableNames) or (variableNames is None and sv.causality == 'output'):
+                type = sv.type
+                if type == 'Enumeration':
+                    type = 'Integer' if modelDescription.fmiVersion in {'1.0', '2.0'} else 'Int32'
+                names, vrs, shapes, n_values, getter = self.info.get(type, ([], [], [], 0, getattr(self.fmu, 'get' + type)))
+                names.append(sv.name)
+                vrs.append(sv.valueReference)
+                shapes.append(sv.dimensions)
+                n_values += np.prod(sv.dimensions) if sv.dimensions else 1
+                self.info[type] = names, vrs, shapes, n_values, getter
 
-                if sv.type == 'Real':
-                    real_names.append(sv.name)
-                    self.real_vrs.append(sv.valueReference)
-                elif sv.type in ['Integer', 'Enumeration']:
-                    integer_names.append(sv.name)
-                    self.integer_vrs.append(sv.valueReference)
-                elif sv.type == 'Boolean':
-                    boolean_names.append(sv.name)
-                    self.boolean_vrs.append(sv.valueReference)
-                else:
-                    pass  # skip String variables
+        # create the columns for the NumPy array
+        if modelDescription.fmiVersion in ['1.0', '2.0']:
+            types = [('Real', np.float64), ('Integer', np.int32), ('Boolean', np.bool_)]
+        else:
+            types = [('Float64', np.float64), ('Int32', np.int32), ('UInt64', np.uint64), ('Boolean', np.bool_)]
 
-        self.cols += zip(real_names, [np.float64] * len(real_names))
-        self.cols += zip(integer_names, [np.int32] * len(integer_names))
-        self.cols += zip(boolean_names, [np.bool_] * len(boolean_names))
+        # collect the columns
+        for t, dt in types:
+            if t in self.info:
+                self.types.append(t)
+                names, _, shapes, _, _ = self.info[t]
+                self.cols += zip(names, [dt] * len(names), shapes)
+
+        # strip the shape for scalars
+        self.cols = [(n, t) if s is None else (n, t, s) for n, t, s in self.cols]
+
+    @staticmethod
+    def _append_reshaped(row, values, shapes):
+        i = 0
+        for d in shapes:
+            if d:
+                s = np.prod(d)
+                value = np.array(values[i:i + s]).reshape(d)
+                i += s
+            else:
+                value = values[i]
+                i += 1
+            row.append(value)
 
     def sample(self, time, force=False):
         """ Record the variables """
@@ -77,14 +92,13 @@ class Recorder(object):
 
         row = [time]
 
-        if self.real_vrs:
-            row += self.fmu.getReal(vr=self.real_vrs)
-
-        if self.integer_vrs:
-            row += self.fmu.getInteger(vr=self.integer_vrs)
-
-        if self.boolean_vrs:
-            row += self.fmu.getBoolean(vr=self.boolean_vrs)
+        for t in self.types:
+            names, vrs, shapes, nValues, getter = self.info[t]
+            if self.modelDescription.fmiVersion in ['1.0', '2.0']:
+                values = getter(vr=vrs)
+            else:
+                values = getter(vr=vrs, nValues=nValues)
+            self._append_reshaped(row, values, shapes)
 
         self.rows.append(tuple(row))
 
@@ -134,6 +148,7 @@ class Input(object):
         self.t_events = Input.findEvents(signals, modelDescription)
 
         is_fmi1 = isinstance(fmu, _FMU1)
+        is_fmi2 = isinstance(fmu, _FMU2)
 
         setters = dict()
 
@@ -142,10 +157,14 @@ class Input(object):
             setters['Real']    = (fmu.fmi1SetReal,    fmi1Real)
             setters['Integer'] = (fmu.fmi1SetInteger, fmi1Integer)
             setters['Boolean'] = (fmu.fmi1SetBoolean, c_int8)
-        else:
+        elif is_fmi2:
             setters['Real']    = (fmu.fmi2SetReal,    fmi2Real)
             setters['Integer'] = (fmu.fmi2SetInteger, fmi2Integer)
             setters['Boolean'] = (fmu.fmi2SetBoolean, fmi2Boolean)
+        else:
+            setters['Float64'] = (fmu.fmi3SetFloat64, fmi3.fmi3Float64)
+            setters['Int32']   = (fmu.fmi3SetInt32,   fmi3.fmi3Int32)
+            setters['Boolean'] = (fmu.fmi3SetBoolean, fmi3.fmi3Boolean)
 
         from collections import defaultdict
 
@@ -164,7 +183,7 @@ class Input(object):
                 print("Warning: missing input for " + sv.name)
                 continue
 
-            if sv.type == 'Real' and sv.variability not in ['discrete', 'tunable']:
+            if sv.type in {'Float32', 'Float64', 'Real'} and sv.variability not in ['discrete', 'tunable']:
                 continuous_inputs[sv.type].append((sv.valueReference, sv.name))
             else:
                 # use the same table for Integer and Enumeration
@@ -194,11 +213,16 @@ class Input(object):
         if self.t is None:
             return
 
+        is_fmi3 = isinstance(self.fmu, fmi3._FMU3)
+
         # continuous
         if continuous:
             for vrs, values, table, setter in self.continuous:
                 values[:] = self.interpolate(time=time, t=self.t, table=table, discrete=False, after_event=after_event)
-                setter(self.fmu.component, vrs, len(vrs), values)
+                if is_fmi3:
+                    setter(self.fmu.component, vrs, len(vrs), values, len(values))
+                else:
+                    setter(self.fmu.component, vrs, len(vrs), values)
 
         # discrete
         if discrete:
@@ -209,7 +233,10 @@ class Input(object):
                     # special treatment for fmi1Boolean
                     setter(self.fmu.component, vrs, len(vrs), cast(values, POINTER(c_char)))
                 else:
-                    setter(self.fmu.component, vrs, len(vrs), values)
+                    if is_fmi3:
+                        setter(self.fmu.component, vrs, len(vrs), values, len(values))
+                    else:
+                        setter(self.fmu.component, vrs, len(vrs), values)
 
     def nextEvent(self, time):
         """ Get the next input event """
@@ -308,20 +335,28 @@ def apply_start_values(fmu, model_description, start_values, apply_default_start
 
         vr = variable.valueReference
 
-        if variable.type == 'Real':
-            fmu.setReal([vr], [float(value)])
-        elif variable.type in ['Integer', 'Enumeration']:
-            fmu.setInteger([vr], [int(value)])
-        elif variable.type == 'Boolean':
+        # get the setter function
+        if variable.type == 'Enumeration':
+            if model_description.fmiVersion in {'1.0', '2.0'}:
+                setter = getattr(fmu, 'setInteger')
+            else:
+                setter = getattr(fmu, 'setInt32')
+        else:
+            setter = getattr(fmu, 'set' + variable.type)
+
+        # convert Boolean values
+        if variable.type == 'Boolean':
             if isinstance(value, str):
                 if value.lower() not in ['true', 'false']:
                     raise Exception('The start value "%s" for variable "%s" could not be converted to Boolean' %
                                     (value, variable.name))
                 else:
                     value = value.lower() == 'true'
-            fmu.setBoolean([vr], [bool(value)])
-        elif variable.type == 'String':
-            fmu.setString([vr], [value])
+
+        # convert the type
+        value = variable._python_type(value)
+
+        setter([vr], [value])
 
     if len(start_values) > 0:
         raise Exception("The start values for the following variables could not be set because they don't exist: " +
@@ -493,11 +528,16 @@ def simulate_fmu(filename,
         callbacks.allocateMemory = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
         callbacks.freeMemory = fmi1CallbackFreeMemoryTYPE(freeMemory)
         callbacks.stepFinished = None
-    else:
+    elif model_description.fmiVersion == '2.0':
         callbacks = fmi2CallbackFunctions()
         callbacks.logger = fmi2CallbackLoggerTYPE(logger)
         callbacks.allocateMemory = fmi2CallbackAllocateMemoryTYPE(allocateMemory)
         callbacks.freeMemory = fmi2CallbackFreeMemoryTYPE(freeMemory)
+    else:
+        callbacks = fmi3.fmi3CallbackFunctions()
+        callbacks.logger = fmi3.fmi3CallbackLoggerTYPE(logger)
+        callbacks.allocateMemory = fmi3.fmi3CallbackAllocateMemoryTYPE(fmi3.allocateMemory)
+        callbacks.freeMemory = fmi3.fmi3CallbackFreeMemoryTYPE(fmi3.freeMemory)
 
     # simulate_fmu the FMU
     if fmi_type == 'ModelExchange' and model_description.modelExchange is not None:
@@ -541,25 +581,29 @@ def simulateME(model_description, fmu_kwargs, start_time, stop_time, solver_name
 
     is_fmi1 = model_description.fmiVersion == '1.0'
 
-    if is_fmi1:
+    if model_description.fmiVersion == '1.0':
         fmu = FMU1Model(**fmu_kwargs)
         fmu.instantiate(functions=callbacks, loggingOn=debug_logging)
         fmu.setTime(time)
-    else:
+    elif model_description.fmiVersion == '2.0':
         fmu = FMU2Model(**fmu_kwargs)
         fmu.instantiate(callbacks=callbacks, loggingOn=debug_logging)
         fmu.setupExperiment(startTime=start_time)
+    else:
+        fmu = fmi3.FMU3Model(**fmu_kwargs)
+        fmu.instantiate(callbacks=callbacks, loggingOn=debug_logging)
+        fmu.setupExperiment(startTime=start_time)
 
     input = Input(fmu, model_description, input_signals)
 
+    apply_start_values(fmu, model_description, start_values, apply_default_start_values)
+
     # initialize
     if is_fmi1:
-        apply_start_values(fmu, model_description, start_values, apply_default_start_values)
         input.apply(time)
         fmu.initialize()
     else:
         fmu.enterInitializationMode()
-        apply_start_values(fmu, model_description, start_values, apply_default_start_values)
         input.apply(time)
         fmu.exitInitializationMode()
 
@@ -728,23 +772,27 @@ def simulateCS(model_description, fmu_kwargs, start_time, stop_time, relative_to
     if model_description.fmiVersion == '1.0':
         fmu = FMU1Slave(**fmu_kwargs)
         fmu.instantiate(functions=callbacks, loggingOn=debug_logging)
-    else:
+    elif model_description.fmiVersion == '2.0':
         fmu = FMU2Slave(**fmu_kwargs)
         fmu.instantiate(callbacks=callbacks, loggingOn=debug_logging)
         fmu.setupExperiment(tolerance=relative_tolerance, startTime=start_time)
+    else:
+        fmu = fmi3.FMU3Slave(**fmu_kwargs)
+        fmu.instantiate(callbacks=callbacks, loggingOn=debug_logging)
+        fmu.setupExperiment(tolerance=relative_tolerance, startTime=start_time)
 
     input = Input(fmu=fmu, modelDescription=model_description, signals=input_signals)
 
     time = start_time
 
+    apply_start_values(fmu, model_description, start_values, apply_default_start_values)
+
     # initialize the model
     if model_description.fmiVersion == '1.0':
-        apply_start_values(fmu, model_description, start_values, apply_default_start_values)
         input.apply(time)
         fmu.initialize()
     else:
         fmu.enterInitializationMode()
-        apply_start_values(fmu, model_description, start_values, apply_default_start_values)
         input.apply(time)
         fmu.exitInitializationMode()
 
